/*******************************************************************************
' Author: Dave Hein
' Version 0.21
' Copyright (c) 2010, 2011
' See end of file for terms of use.
'******************************************************************************/
#include <stdio.h>
#include <string.h>
#include "interp.h"
#include "spinsim.h"

#define ADDR_MASK 0xfffff

extern char *hubram;
extern int32_t memsize;
extern int32_t loopcount;
extern int32_t cycleaccurate;
extern int32_t pin_val;

extern FILE *tracefile;

static char *condnames[16] = {
    "if_never    ", "if_nz_and_nc", "if_z_and_nc ", "if_nc       ",
    "if_nz_and_c ", "if_nz       ", "if_z_ne_c   ", "if_nz_or_nc ",
    "if_z_and_c  ", "if_z_eq_c   ", "if_z        ", "if_z_or_nc  ",
    "if_c        ", "if_nz_or_c  ", "if_z_or_c   ", "            "};

static char *group1[] = {
    "ror", "rol", "shr", "shl", "rcr", "rcl", "sar", "sal",
    "add", "addx", "adds", "addsx", "sub", "subx", "subs", "subsx",
    "cmp", "cmpx", "cmps", "cmpsx", "cmpr", "cmpm", "subr", "cmpsub",
    "min", "max", "mins", "maxs", "sumc", "sumnc", "sumz", "sumnz",
    "isob", "notb", "clrb", "setb", "setbc", "setbnc", "setbz", "setbnz",
    "andn", "and", "or", "xor", "muxc", "muxnc", "muxz", "muxnz",
    "mov", "not", "abs", "neg", "negc", "negnc", "negz", "negnz",
    "alti", "decod", "topone", "botone", "incmod", "decmod", "mul", "muls",
    "setnibn", "setnibn", "getnibn", "getnibn", "rolnibn", "rolnibn", "setbytn", "getbytn"};

static char *group2[] = {
    "rolbytn", "rolbytn", "rolbytn", "rolbytn", "setwrdn", "setwrdn", "getwrdn", "getwrdn",
    "rolwrdn", "rolwrdn", "setbyts", "movbyts", "splitb", "mergeb", "splitw", "mergew",
    "seussf", "seussr", "rgbsqz", "rgbexp", "rev", "seti", "setd", "sets",
    "djz", "djnz", "djs", "djns", "tjz", "tjnz", "tjs", "tjns",
    "testn", "testn", "testn", "testn", "test", "test", "test", "test",
    "anyb", "anyb", "anyb", "anyb", "testb", "testb", "testb", "testb",
    "addct1", "addct2", "addct3", "wmlong", "calld", "calld", "calld", "calld",
    "msgin", "msgin", "msgin", "msgin", "rdlut", "rdlut", "rdlut", "rdlut",
    "rdbyte", "rdbyte", "rdbyte", "rdbyte", "rdword", "rdword", "rdword", "rdword",
    "rdlong", "rdlong", "rdlong", "rdlong", "addpix", "mulpix", "blnpix", "mixpix",
    "invalid", "invalid", "invalid", "invalid", "invalid", "invalid", "invalid", "invalid",
    "setpae", "setpae", "setpan", "setpan", "setpbe", "setpbe", "setpbn", "setpbn",
    "jp", "jp", "jnp", "jnp", "msgout", "msgout", "wrlut", "wrlut",
    "wrbyte", "wrbyte", "wrword", "wrword", "wrlong", "wrlong", "rdfast", "rdfast",
    "wrfast", "wrfast", "fblock", "fblock", "xinit", "xinit", "xzero", "xzero",
    "xcont", "xcont", "rep", "rep", "coginit", "coginit", "coginit", "coginit",
    "qmul", "qmul", "qdiv", "qdiv", "qfrac", "qfrac", "qsqrt", "qsqrt",
    "qrotate", "qrotate", "qvector", "qvector"};

static char *group3[] = {
    "jmp", "call", "calla", "callb", "calld", "calld", "calld", "calld",
    "loc", "loc", "loc", "loc", "augs", "augs", "augs", "augs",
    "augd", "augd", "augd", "augd"};

static char *group4[] = {
    "clkset", "cogid", "invalid", "cogstop", "locknew", "lockret", "lockclr", "lockset",
    "invalid", "invalid", "invalid", "invalid", "invalid", "invalid", "qlog", "qexp",
    "rfbyte", "rfword", "rflong", "wfbyte", "wfword", "wflong", "setq", "setq2",
    "getqx", "getqy", "getct", "getrnd", "setdacs", "setxfrq", "getxcos", "getxsin",
    "setedg", "setrdl", "setwrl", "sethlk", "pollint", "setint1", "setint2", "setint3",
    "waitx", "setcz", "push", "pop", "jmp", "call", "calla", "callb", 
    "jmprel", "ret", "reta", "retb", "getptr", "getint", "setbrk", "invalid",
    "setcy", "setci", "setcq", "setcfrq", "setcmod", "setpix", "setpiv", "invalid"};

static char *group5[] = {
    "pollint", "pollct1", "pollct2", "pollct3", "pollpat", "polledg", "pollrdl", "pollwrl",
    "pollhlk", "pollxmt", "pollxfi", "pollxro", "pollfbw", "pollrle", "invalid", "invalid",
    "waitint", "waitct1", "waitct2", "waitct3", "waitpat", "waitedg", "waitrdl", "waitwrl",
    "waithlk", "waitxmt", "waitxfi", "waitxro", "waitfbw", "waitrle", "invalid", "invalid",
    "allowi", "stalli"};

static char *group6[] = {"alti", "altr", "altd", "alts"};

char *GetOpname2(unsigned int instr, int *pczi)
{
    int opcode = (instr >> 21) & 0x7f;
    static char name[8];

    //printf("opcode = $%2.2x, czi = %d\n", opcode, czi);

    if (opcode <= 0x47)
    {
        if (opcode == 0x38)
        {
            strcpy(name, group6[(instr >> 19) & 3]);
	    *pczi = 1;
        }
        else
        {
            strcpy(name, group1[opcode]);

            if (opcode >= 0x40)
	        *pczi = 1;
            else
	        *pczi = 7;
        }
    }
    else if (opcode <= 0x6a)
    {
        strcpy(name, group2[((instr >> 19) & 0x1ff) - (0x48 << 2)]);

        if (opcode >= 0x50 && opcode <= 0x5a)
            *pczi = 7;
        else if (opcode >= 0x5c && opcode <= 0x5f)
            *pczi = 5;
        else if (opcode >= 0x61 && opcode <= 0x63)
            *pczi = 7;
        else
	    *pczi = 1;
    }
    else if (opcode == 0x6b)
    {
        if ((instr & 0x3f) == 0x24)
        {
            int temp = (instr >> 9) & 0x1ff;
            if (temp < 34)
                strcpy(name, group5[temp]);
            else
                strcpy(name, "invalid");
        }
        else
            strcpy(name, group4[instr & 0x3f]);
	*pczi = 7;
    }
    else
    {
        strcpy(name, group3[opcode - 0x6c]);
        if (opcode >= 0x78)
	    *pczi = 0;
        else
	    *pczi = 1;
    }

    return name;
}

void StartPasmCog2(PasmVarsT *pasmvars, int32_t par, int32_t addr, int32_t cogid, int32_t hubexec)
{
    int32_t i;

    //printf("\nStartPasmCog2: %8.8x, %8.8x, %d, %d\n", par, addr, cogid, hubexec);
    par &= ADDR_MASK;
    addr &= ADDR_MASK;
    pasmvars->waitflag = 0;
    pasmvars->cflag = 0;
    pasmvars->zflag = 0;
    pasmvars->cogid = cogid;
    pasmvars->state = 5;
    pasmvars->ptra = par;
    pasmvars->ptrb = addr;
    pasmvars->ptrx = 0;
    pasmvars->ptry = 0;
    pasmvars->inda = 0x1f2;
    pasmvars->indatop = 0;
    pasmvars->indabot = 0;
    pasmvars->indb = 0x1f3;
    pasmvars->indbtop = 0;
    pasmvars->indbbot = 0;
    pasmvars->repcnt = 0;
    pasmvars->repbot = 0;
    pasmvars->reptop = 0;
    pasmvars->repforever = 0;
    pasmvars->instruct1 = 0;
    pasmvars->instruct2 = 0;
    pasmvars->instruct3 = 0;
    pasmvars->instruct4 = 0;
    pasmvars->pc1 = INVALIDATE_INSTR;
    pasmvars->pc2 = INVALIDATE_INSTR;
    pasmvars->pc3 = INVALIDATE_INSTR;
    pasmvars->pc4 = INVALIDATE_INSTR;
    pasmvars->retptr = 0;
    pasmvars->acca = 0;
    pasmvars->accb = 0;
    pasmvars->cordic_count = 0;
    pasmvars->cordic_depth = 0;
    pasmvars->qxposted = 0;
    pasmvars->qyposted = 0;
    pasmvars->breakpnt = -1;
    pasmvars->augsflag = 0;
    pasmvars->augsvalue = 0;
    pasmvars->augdflag = 0;
    pasmvars->augdvalue = 0;
    pasmvars->altsflag = 0;
    pasmvars->altsvalue = 0;
    pasmvars->altdflag = 0;
    pasmvars->altdvalue = 0;
    pasmvars->altrflag = 0;
    pasmvars->altrvalue = 0;
    pasmvars->altiflag = 0;
    pasmvars->altivalue = 0;
    pasmvars->prefetch = 0;
    pasmvars->serina.mode = 0;
    pasmvars->serinb.mode = 0;
    pasmvars->serouta.mode = 0;
    pasmvars->seroutb.mode = 0;
    pasmvars->phase = 0;
    pasmvars->rwrep = 0;
    pasmvars->str_fifo_work_flag = 0;
    pasmvars->str_fifo_addr0 = 0;
    pasmvars->str_fifo_addr1 = 0;
    pasmvars->str_fifo_head_addr = 0;
    pasmvars->str_fifo_tail_addr = 0;
    pasmvars->str_fifo_rindex = 0;
    pasmvars->str_fifo_windex = 0;
    pasmvars->str_fifo_mode = 0;
    pasmvars->str_fifo_buffer[0] = 0;
    pasmvars->cntreg1 = 0;
    pasmvars->cntreg2 = 0;
    pasmvars->cntreg3 = 0;
    pasmvars->intflags = 0;
    pasmvars->intstate = 0;
    pasmvars->qreg = 0;
    pasmvars->memflag = 0;
    pasmvars->intenable1 = 0;
    pasmvars->intenable2 = 0;
    pasmvars->intenable3 = 0;
    pasmvars->pinpatmode = 0;
    pasmvars->pinpatmask = 0;
    pasmvars->pinpattern = 0;
    pasmvars->pinedge = 0;
    pasmvars->lockedge = 0;
    pasmvars->rdl_mask = 0;
    pasmvars->wrl_mask = 0;

    if (!hubexec)
    {
        for (i = 0; i < 0x1f4; i++)
        {
            pasmvars->mem[i] = LONG(addr);
            addr += 4;
        }
        pasmvars->pc = 0;
    }
    else
        pasmvars->pc = addr;
    for (i = 0x1f4; i < 512; i++) pasmvars->mem[i] = 0;
}

void DebugPasmInstruction2(PasmVarsT *pasmvars)
{
    int32_t i;
    int32_t cflag = pasmvars->cflag;
    int32_t zflag = pasmvars->zflag;
    int32_t instruct, pc, cond, xflag;
    int32_t opcode, czi;
    int32_t srcaddr, dstaddr;
    char *wzstr = "";
    char *wcstr = "";
    char opstr[20];
    char *istr[4] = {" ", "#", "@", "&"};
    char *xstr[11] = {" ", "X", "I", "H", "C", "P", "W", "F", "w", "E", "?"};
    int czi_mask;
    int32_t sflag = 0;
    int32_t dflag = 0;

    // Fetch the instruction
    pc = pasmvars->pc2;
    instruct = pasmvars->instruct2;
    cond = (instruct >> 28) & 15;

    // Extract parameters from the instruction
    opcode = (instruct >> 21) & 127;
    srcaddr = instruct & 511;
    dstaddr = (instruct >> 9) & 511;
    czi = (instruct >> 18) & 7;

    // Decode the immediate flags for the source and destination fields
    if ((czi & 1) && opcode <= 0x6a)
    {
        if (!pasmvars->augsflag && (instruct & 0x100) && ((opcode >= 0x58 &&
            opcode <= 0x5a) || opcode == 0x62 || (opcode == 0x63 && !(czi&4))))
            sflag = 3;
        else if ((opcode >= 0x4e && opcode <= 0x4f) || opcode == 0x55 || opcode == 0x60)
            sflag = 2;
        else
            sflag = 1;
    }

    dflag = ((czi & 1) && (opcode == 0x6b)) ||
            ((czi & 2) && (opcode >= 0x5d && opcode <= 0x6a));

    // Check for extended address instructions
    if (opcode >= 0x6c)
    {
        if (opcode < 0x78)
        {
            srcaddr = instruct & 0xfffff;
            sflag = (czi >> 2) + 1;
            if (opcode >= 0x78)
                dstaddr = 0x1f6 + (opcode & 3);
        }
        else
        {
            srcaddr = instruct & 0x7fffff;
            sflag = 1;
        }
    }

    xflag = ((cond >> ((cflag << 1) | zflag)) & 1);
    xflag ^= 1;

    // Check if the instruction is invalidated in the pipeline
    if (pc & INVALIDATE_INSTR)
    {
	pc &= ~INVALIDATE_INSTR;
	xflag = 2;
    }

    if (pasmvars->waitflag && pasmvars->waitmode == WAIT_CACHE) xflag = 4;
    else if (xflag == 4) xflag = 5;

    if (pasmvars->waitflag)
    {
	if (pasmvars->waitmode == WAIT_CACHE)
	    xflag = 4;
	else if (pasmvars->waitmode == WAIT_CNT)
	    xflag = 6;
	else if (pasmvars->waitmode == WAIT_PIN)
	    xflag = 5;
	else if (pasmvars->waitmode == WAIT_HUB)
	    xflag = 3;
	else if (pasmvars->waitmode == WAIT_FLAG)
	    xflag = 8;
	else if (pasmvars->waitmode == WAIT_CORDIC)
	    xflag = 9;
	else
	    xflag = 10;
    }
    else if (pasmvars->phase == 0)
        xflag = 7;

    strcpy(opstr, GetOpname2(instruct, &czi_mask));

    czi &= czi_mask;
    if (czi & 4) wzstr = " wc";
    if (czi & 2) wcstr = " wz";

    i = strlen(opstr);
    while (i < 7) opstr[i++] = ' ';
    opstr[i] = 0;

    // Check for NOP
    if (!instruct)
    {
	cond = 15;
	if (xflag == 1) xflag = 0;
	strcpy(opstr, "nop    ");
    }

    if (pasmvars->printflag - 2 < xflag && xflag != 0)
    {
	pasmvars->printflag = 0;
	return;
    }

    fprintf(tracefile, "Cog %d: %8.8x ", pasmvars->cogid, loopcount);
    fprintf(tracefile, "%4.4x %8.8x %s %s %s %s%3.3x, %s%3.3x%s%s", pc,
        instruct, xstr[xflag], condnames[cond], opstr, istr[dflag], dstaddr,
	istr[sflag], srcaddr, wzstr, wcstr);
}
/*
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
*/
